
open "complement_h"
open "gensym_h"

(* Say we have
 P(t1) :- body1  ...  P(tn) :- bodyn
 that is, P(x) iff (exists ... x=t1 and body1) \/ ... \/ (exists ... x=tn and bodyn).
 (where ... in each case is the list of variables free in ti or bodyi.)
 Then not P(x) iff (forall ... x=t1 => not body1) and ... and (forall ... x=tn => not bodyn).
 This works only if each ti is linear and the free variables of ti contain those of bodyi.
 Then the foralls can be replaced by existentials.
 That is, not P(x) is equivalent to a conjunction
 of expressions (i) forall ... . x=ti => (not bodyi1 and ... and not bodyiki)
 where:
 . there is no x both equal to ti and tj, i<>j;
 . every x is an instance of some ti
 We call the collection of terms ti a *cover*.
 Then we generate the clauses
 not-P(ti) :- not bodyi1 and ... and not bodyiki.

 Given P(x,y), can we encode forall y.P(x,y) as some Q(x)?
 In general, I guess not.  However, we can encode upper approximations Q.

 1. For example, we code:
 Q(x) :- P(x,y)
 This recognizes Q(x) as soon as there exists y such that P(x,y).
 This is correct provided forall y.P(x,y) => exists y.P(x,y),
 i.e., provided the sort of y is not empty.

 2. Or enumerate all constructors of the sort of y: f1, ..., fn, and code:
 Q (x) :- P (x, f1(...))
 ...
 Q (x) :- P (x, fn(...))

 3. Or look at the definition of P.
 P(t1,u1) :- body1
 ...
 P(tn,un) :- bodyn
 Then assume (t1,u1) ... (tn,un) form a cover,
 forall y.P(x,y) <=> forall x_{u1} . body1 and ... and forall x_{un} . bodyn
 where x_{ui} is the set of free variables of ui.
 If each x_{ui} is linear in bodyi, we can recurse.
 Otherwise, linearize bodyi wrt : it will then recognize more cases, hence
 we'll get an upper approximation.
*)

(* covers are stack machines trying to recognize heads of clauses. *)
datatype cover =
	 C_END of definite_clause list
	 (* return list of clauses with same head.
	  Stack is empty. *)
       | C_CASE of string -m> cover
	 (* f(t1,...,tn) --- tn ... t1
	  if term on top of stack matches symbol f,
	  then continue with corresponding cover. *)
       | C_POP of string * sort * cover
	 (* t --- *)
	 ;

val empty_cover = C_END nil;

exception CoverSort of string;
exception CoverFsig of string;
exception CoverWrongArity;
exception UnionCoverCase of string;
exception UCover;
exception CoverFVarInBody of string;

fun tsubst_exact ren =
    let fun ts (V x) =
	    if x inset ren
		then V (?ren x)
	    else raise CoverFVarInBody x
	  | ts (f $ l) =
	    f $ [ts t | t in list l]
    in
	ts
    end;

fun dsubst_exact ren =
    let val ts = tsubst_exact ren
	fun asb (ATOM (P, l)) =
	    ATOM (P, [ts t | t in list l])
    in
	fn DCLAUSE (name, env, h, body) =>
	   DCLAUSE (name,
		    {y => srt
		    | x => srt in map ren <| env
		      val y = ?ren x},
		      asb h,
		      [asb a | a in list body])
    end;

fun cover_of_clause (invc, fsig, Psort, dcl as DCLAUSE (name, env, ATOM (_, l), body)) =
    let fun c_pops (nil, nil, _) = C_END nil
	  | c_pops (V x::rest, srt::rests, ren) =
	    if x inset ren
		then raise NonLinearHead
	    else C_POP (x, srt, c_pops (rest, rests, ren ++ {x => x}))
	fun cover_head (nil, nil, ren) =
	    let val dcl' = dsubst_exact ren dcl
	    in
		C_END [dcl']
	    end
	  | cover_head (V x::rest, srt::rests, ren)
	    if x inset ren
		then raise NonLinearHead
	    else let val y = gensym "__x"
		 in
		     C_POP (y, srt, cover_head (rest, rests, ren ++ {x => y}))
		 end
	  | cover_head ((f $ l)::rest, src::rests, ren)
	    let val constr = ?invc srt handle MapGet => raise CoverSort srt
		val (_, sl) = ?fsig f handle MapGet => raise CoverFsig f
	    in
		C_CASE ({g =>
			  let val (_, sl') = ?fsig g handle MapGet => raise CoverFsig g
			      val xl = [V (gensym "__x") | _ in list sl']
			  in
			      c_pops (xl, sl', ren)
			  end
			| g in set constr} ++
			  {f => cover_head (l @ rest, sl @ rests, ren)})
	    end
	  | cover_head _ = raise CoverWrongArity
    in
	cover_head (l, Psort, {})
    end;

fun subst_of_cover_sigma (nil, sigma) = sigma
  | subst_of_cover_sigma ((x, t)::rest, sigma) =
    subst_of_cover_sigma (rest, sigma ++ {x => tsubst sigma t});

fun cover_subst_dclauses (dcls, subst, newenv) =
    let val sigma = subst_of_cover_sigma (subst, {})
	val asubst = atom_subst sigma
    in
	[DCLAUSE (name, (sigma <-| env) ++ newenv,
		  asubst h,
		  [asubst a | a in list body ])
	| DCLAUSE (name, env, h, body) in list dcls]
    end;


fun union_cover fsig =
    let fun upops (nil, nil, c) = c
	  | upops (x::rest, s::rests, c) =
	    C_POP (x, s, upops (rest, rests, c))
	fun ucover (C_END l, C_END l', sigma, sigma', env, env') =
	    C_END (cover_subst_dclauses (l, sigma, env) @
		   cover_subst_dclauses (l', sigma', env'))
	  | ucover (C_CASE m, C_CASE m', sigma, sigma', env, env') =
	    {f => ucover (c, c', sigma, sigma', env, env')
	    | f => c in map m
		val c' = ?m' f handle MapGet => raise UnionCoverCase f}
	  | ucover (C_POP (x, s, c), C_POP (x', _, c'), sigma, sigma', env, env') =
	    C_POP (x, s, ucover (c, c', sigma, (x', V x)::sigma', env, env'))
	  | ucover (C_CASE m, C_POP (x', _, c'), sigma, sigma', env, env') =
	    C_CASE {f => let val (_, sl) = ?fsig f handle MapGet => raise CoverFsig f
			     val xl = [V (gensym "__x") | _ in list sl]
			 in
			     ucover (c, upops (xl, sl, c'), sigma, (x', f $ xl)::sigma',
				     env,
				     env' ++ {x' => s' || V x' in list xl and s' in list sl})
			 end
		   | f => c in map m}
	  | ucover (C_POP (x, c), C_CASE m', sigma, sigma', env, env') =
	    C_CASE {f => let val (_, sl) = ?fsig f handle MapGet => raise CoverFsig f
			     val xl = [V (gensym "__x") | _ in list sl]
			 in
			     ucover (upops (xl, sl, c), c', (x, f $ xl)::sigma, sigma',
				     env ++ {x => s || V x in list xl and s in list sl},
				     env')
			 end
		   | f => c' in map m'}
	  | ucover _ = raise UCover
    in
	fn (c, c') => ucover (c, c', {}, {}, {}, {})
    end;

fun get_cover (invc, fsig, Psort) =
    let val ucover = union_cover fsig
	fun acc_cover (nil, acc) = acc
	  | acc_cover (dcl::rest, acc) =
	    acc_cover (rest, ucover (acc, cover_of_clause (invc, fsig, Psort, dcl)))
    in
	fn dcls => acc_cover (dcls, empty_cover Psort)
    end;

exception ComplementFalse;

fun complement_cover (psigr, not_pred, add) =
    let memofun bvars body = union {avars a | a in list body}
	fun not_body (_, nil) = raise ComplementFalse
	  | not_body (_, [ATOM (Q, l)]) = ATOM (not_pred Q, l)
	  | not_body (env, body) =
	    let val notB = gensym "_notB"
		val f as |[put, convert, tell, seek, ...]| = outstring notB
		val pos = (put ":"; tell ())
		val env' = bvars body <| env
		val notBargs = [V x | x in set env']
		val notBsort = [s | _ => s in map env']
		val i = ref 0
		val h' = ATOM (notB, notBargs)
	    in
		psigr := !psigr ++ {notB => notBsort};
		iterate
		  (seek pos;
		   inc i;
		   print f (pack (!i));
		   add (DCLAUSE (convert (), env', h', [ATOM (not_pred Q, l)])))
		| ATOM (Q, l) in list body
		end;
		h'
	    end
	fun ccover (C_END dcls) =
	    (* These are clauses of the form P(t) :- body1 ... P(t) :- bodyn
	     with the same head t.
	     Generate not-P(t) :- not body1 and ... and not bodyn.
	     *)
	    let val f as |[put, convert, ...]| = outstring "not-["
		val delim = ref ""
		val notPenv = ref {}
		val notBody = []
	    in
		iterate
		  (put (!delim);
		   delim := ",";
		   put name;
		   notPenv := !notPenv ++ (bvars body <| env))
		| DCLAUSE (name, env, _, body) in list dcls
		end;
		put "]";
		let val notBody = [not_body (env, body)
				  | DCLAUSE (_, env, _, body) in list dcls]
		in
		    add (DCLAUSE (convert (), !notPenv,
				  h, notBody))
		end handle ComplementFalse => ()
	    end
    in
	ccover
    end;

fun complement_predicate (PROG (predsig, prog), fsig, P) =
    let val psigr = ref predsig
	val pr = ref prog
	val constrs = {f => srt
		      | f => (srt, _) in map fsig}
	val invc = invrel constrs
	val worklist = ref [P]
	memofun not_pred Q = let val f as |[put, convert, ...]| = outstring "not-["
			     in
				 put Q;
				 put "]";
				 convert ()
			     end
    in
	while not (null (!worklist)) do
	    let val Q::rest = !worklist
		val notQ = not_pred Q
	    in
		worklist := rest;
		if notQ inset !pr
		    then (* this one already computed *)
			()
		else let val dcls = ?(!pr) Q handle MapGet => nil
			 val Qsort = ?(!psigr) Q handle MapGet => raise NoPredSig Q
			 val cover = get_cover (invc, fsig, Qsort) dcls
		     in
			 psigr := !psigr ++ {notQ => Qsort};
			 
		     end
	    end
    end;

