%{

#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>

extern char *yytext;
extern char string_buf[];

extern void yyerr(char *msg, char *more); /* in spec.l */
extern void yywarn(char *msg, char *more); /* in spec.l */
#define YYERROR_VERBOSE
static struct cell *theCommand = NULL;
extern int theSpecLineno;

jmp_buf errbuf;
int yyparse (void);
extern int yylex (void);

#define X_FUNC 0
#define X_INT 1
#define X_STRING 2
#define X_CONS 3

typedef struct cell {
  unsigned int type:2;
  unsigned int n:30;
  union {
    struct func { struct cell *(*fun) (struct cell *, struct cell *); struct cell *env; } f;
    int i;
    struct str { struct str *next; char *text; } s;
    struct cons { struct cell *car, *cdr; } c;
  } what;
} cell;

#define NCELLS 1024

struct cellblock {
  struct cellblock *next;
  cell c[NCELLS];
};

void *Malloc(size_t size)
{
  void *p;

  p = malloc(size);
  if (p==NULL)
    {
      fprintf (stderr, "Sorry, cannot allocate %d bytes.\n", size);
      exit(10);
    }
  return p;
}

static void yyerror(char *msg)
  {
    int len;
    char *more, *s;

    len = strlen(yytext) + 48;
    if (theCommand!=NULL)
      len += strlen (theCommand->what.s.text) + 16;
    s = more = (char *)Malloc(len+16);
    more[0] = 0;
    if (theCommand!=NULL)
      {
	sprintf (more, " in '%s'", theCommand->what.s.text);
	s = more + strlen (more);
      }
    sprintf (s, ", line %d", theSpecLineno);
    s += strlen (s);
    if (yytext[0])
      sprintf(s, " (at '%s')", yytext);
    yyerr (msg, more);
    /*
    free(more);
    */
  }

static struct cellblock *theCellBlocks = NULL;
static cell *theFreeCells = NULL;
static struct str *theStrs = NULL;

static cell *theReturn;

struct yyltype;

static void end_parse (void)
  {
    struct cellblock *cb;
    struct str *str;

    for (str = theStrs; str!=NULL; str = str->next)
      free (str->text);
    theStrs = NULL;
    while (theCellBlocks!=NULL)
      {
	cb = theCellBlocks->next;
	free (cb);
	theCellBlocks = cb;
      }
  }

static cell *newcell (void)
  {
    cell *c, *cc;
    struct cellblock *cb;
    int i;

    c = theFreeCells;
    if (c!=NULL)
      {
	theFreeCells = c->what.c.cdr;
	return c;
      }
    cb = Malloc (sizeof (struct cellblock));
    cb->next = theCellBlocks;
    theCellBlocks = cb;
    for (i=0, cc=cb->c; i<NCELLS; i++, cc++)
      {
	cc->what.c.cdr = c;
	c = cc;
      }
    theFreeCells = c->what.c.cdr;
    return c;
  }

static cell *newfun (struct cell *(*fun)(struct cell *, struct cell *),
		     struct cell *env)
  {
    cell *c = newcell ();

    c->type = X_FUNC;
    c->n = 0;
    c->what.f.fun = fun;
    c->what.f.env = env;
    return c;
  }

static cell *newint (int n)
  {
    cell *c = newcell ();

    c->type = X_INT;
    c->n = 0;
    c->what.i = n;
    return c;
  }

static cell *newstring (char *text)
  {
    cell *c = newcell ();

    c->type = X_STRING;
    c->n = 0;
    c->what.s.next = theStrs;
    c->what.s.text = Malloc (strlen (text)+1);
    strcpy (c->what.s.text, text);
    theStrs = &c->what.s;
    return c;
  }

static cell *cons (cell *car, cell *cdr)
  {
    cell *c = newcell ();

    c->type = X_CONS;
    c->n = 0;
    c->what.c.car = car;
    c->what.c.cdr = cdr;
    return c;
  }

static char *tags[] = {
#define T_DUMMY 0
  "",
#define T_LINE 1
  "line",
#define T_SPEC 2
  "spec",
#define T_DECLARE 3
  "declare",
#define T_SESSIONS 4
  "sessions",
#define T_PROPS 5
  "props",
#define T_TYPE 6
  "type",
#define T_FUNCTION 7
  "function",
#define T_ONE_WAY_FUNCTION 8
  "one-way-function",
#define T_UNIT 9
  "unit",
#define T_PRODUCT 10
  "product",
#define T_ALIAS 11
  "alias",
#define T_KNOWS 12
  "knows",
#define T_INTRUDER 13
  "intruder",
#define T_AXIOM 14
  "axiom",
#define T_STOP 15
  "stop",
#define T_COMM 16
  "comm",
#define T_TST_EQUAL 17
  "equals",
#define T_ASSIGN 18
  "assign",
#define T_SWITCH 19
  "switch",
#define T_CASE 20
  "case",
#define T_TUPLE 21
  "tuple",
#define T_CRYPT 22
  "crypt",
#define T_PERCENT 23
  "percent",
#define T_APP 24
  "apply",
#define T_VANILLA 25
  "vanilla",
#define T_SESSION 26
  "session",
#define T_ROLE 27
  "role",
#define T_CLAIM 28
  "claim",
#define T_KEYPAIR 29
  "keypair",
#define T_SIGN 30
  "sign",
};

static cell *apply (int n, cell *args)
  {
    return cons (newint (n), args);
  }

static cell *newlineinfo (struct yyltype *start, struct yyltype *end, cell *c);

static cell *applylineinfo (cell *lineinfo, cell *c)
  {
    return apply (T_LINE,
		  cons (c, lineinfo->what.c.cdr->what.c.cdr));
  }

static cell *reverse (cell *c, cell *tail)
  {
    while (c!=NULL)
      {
	tail = cons (c->what.c.car, tail);
	c = c->what.c.cdr;
      }
    return tail;
  }

static cell *type_prod (struct yyltype *start, struct yyltype *end, cell *tlst)
  {
    if (tlst==NULL)
      return newlineinfo (start, end, apply (T_UNIT, NULL));
    if (tlst->what.c.cdr==NULL)
      return tlst->what.c.car;
    return newlineinfo (start, end,
			apply (T_PRODUCT,
			       reverse (tlst, NULL)));
  }

#define CALL(F,arg) (*(F)->what.f.fun) (arg, (F)->what.f.env)

static cell *do_identity (cell *c, cell *env)
  {
    return c;
  }

static cell *do_compose (cell *arg, cell *f_and_g)
  {
    cell *f = f_and_g->what.c.car;
    cell *g = f_and_g->what.c.cdr;

    return CALL (f, CALL (g, arg));
  }

static cell *compose (cell *f, cell *g)
  {
    cell *f_and_g = cons (f, g);
    return newfun (do_compose, f_and_g);
  }

static cell *do_comm (cell *k, cell *linfo_and_commrest)
  {
    cell *linfo = linfo_and_commrest->what.c.car;
    cell *commrest = linfo_and_commrest->what.c.cdr;
    return applylineinfo (linfo,
			  apply (T_COMM,
				 cons (k, commrest)));
  }

static cell *do_guard (cell *k, cell *env)
  {
    int eqop = env->what.c.car->what.i;
    cell *linfo = env->what.c.cdr->what.c.car;
    cell *rest = env->what.c.cdr->what.c.cdr;
    return applylineinfo (linfo,
			  apply (eqop,
				 cons (k, rest)));
  }

static cell *do_switch (cell *k, cell *env)
  {
    cell *t = env->what.c.car;
    cell *cases = env->what.c.cdr->what.c.car;
    cell *linfo = env->what.c.cdr->what.c.cdr;
    cell *c, **cp, *onecase;

    c = cons (t, NULL);
    for (cp = &c->what.c.cdr; cases!=NULL; cases = cases->what.c.cdr, cp = &(*cp)->what.c.cdr)
      {
	onecase = cases->what.c.car;
	*cp = cons (CALL (onecase, k), NULL);
      }
    return applylineinfo (linfo, apply (T_SWITCH, c));
  }

static cell *do_case (cell *k, cell *env)
  {
    cell *t = env->what.c.car;
    cell *action = env->what.c.cdr->what.c.car;
    cell *linfo = env->what.c.cdr->what.c.cdr;

    return applylineinfo (linfo,
			  apply (T_CASE,
				 cons (t,
				       cons (CALL (action, k),
					     NULL))));
  }

static cell *mktuple (struct yyltype *start, struct yyltype *end, cell *tlst)
  {
    if (tlst==NULL)
      return newlineinfo (start, end, apply (T_TUPLE, NULL));
    if (tlst->what.c.cdr==NULL)
      return tlst->what.c.car;
    return newlineinfo (start, end,
			apply (T_TUPLE,
			       reverse (tlst, NULL)));
  }

%}

%union {
  struct cell *c;
  int i;
}

%token kw_int kw_id kw_label kw_alias kw_axiom kw_case kw_hash kw_intruder
%token kw_knows kw_session kw_switch kw_arrow kw_tst_equal kw_assign kw_eof
%token kw_string_constant kw_keypair

%start spec
/* type <term>: */
%type <c> spec type term block atomic_term atomic_term_or_ciphertext optional_encryption_algorithm session assignment
%type <c> property id label
/* type <term list>: */
%type <c> declarations declaration type_list non_empty_type_list
%type <c> tuple term_list environment properties
%type <c> assignments ids
/* type <bool>: */
%type <i> optional_hash_keyword
/* type <int>: */
%type <i> equals_or_assign_operator
/* type <(term -> term)>: */
%type <c> messages message case
/* type <(term -> term) list>: */
%type <c> cases

%%

spec : id declarations block environment properties kw_eof
{ $$ = newlineinfo (&@1, &@5,
		    apply (T_SPEC,
			   cons (newlineinfo (&@2, &@2, apply (T_DECLARE, reverse ($2, NULL))),
			   cons (newlineinfo (&@3, &@3, $3),
			   cons (newlineinfo (&@4, &@4, apply (T_SESSIONS, reverse ($4, NULL))),
			   cons (newlineinfo (&@5, &@5, apply (T_PROPS, reverse ($5, NULL))),
				 NULL))))));
 theReturn = $$;
 YYACCEPT; 
}
;

id : kw_id { $$ = newstring (yytext); }
;

/* declarations est une liste inversee de declarations. */
declarations : { $$ = NULL; }
| declarations declaration { $$ = reverse ($2, $1); }
;

declaration : ids ':' type
{
  cell *c, **cp;
  for (c = $1, cp = &$$, *cp=NULL; c!=NULL; c=c->what.c.cdr, cp = &(*cp)->what.c.cdr)
    *cp = cons (newlineinfo (&@1, &@3, apply (T_TYPE,
					      cons (c->what.c.car,
						    cons ($3, NULL)))),
		NULL);
}
| id '(' type_list ')' ':' type optional_hash_keyword
{
  cell *ctype;
  ctype = apply (($7?T_ONE_WAY_FUNCTION:T_FUNCTION),
		 cons (type_prod (&@2, &@4, $3), cons ($6, NULL)));
  ctype = newlineinfo (&@2, &@7, ctype);
  $$ = cons (newlineinfo (&@1, &@7,
			  apply (T_TYPE,
				 cons ($1, cons (ctype, NULL)))),
	     NULL);
}
| kw_keypair id ',' id
{
  $$ = cons (newlineinfo (&@1, &@4, apply (T_KEYPAIR, cons ($2, cons ($4, NULL)))),
	     NULL);
}
| kw_alias id '=' term
{
  $$ = cons (newlineinfo (&@1, &@4,
			  apply (T_ALIAS,
				 cons ($2, cons ($4, NULL)))),
	     NULL);
}
| id kw_knows tuple
{
  cell *c, **cp;

  $$ = reverse ($3, NULL);
  for (c=$$, cp = &$$, *cp=NULL; c!=NULL; c=c->what.c.cdr, cp = &(*cp)->what.c.cdr)
    *cp = cons (newlineinfo (&@1, &@3,
			     apply (T_KNOWS, cons ($1, cons (c->what.c.car, NULL)))),
		NULL);
}
| kw_intruder kw_knows tuple
{
  cell *c, **cp, *ci;

  ci = apply (T_INTRUDER, NULL);
  $$ = reverse ($3, NULL);
  for (c=$$, cp = &$$, *cp=NULL; c!=NULL; c=c->what.c.cdr, cp = &(*cp)->what.c.cdr)
    *cp = cons (newlineinfo (&@1, &@3,
			     apply (T_KNOWS, cons (ci, cons (c->what.c.car, NULL)))),
		NULL);
}
| kw_axiom term '=' term
{
  $$ = cons (newlineinfo (&@1, &@4,
			  apply (T_AXIOM, cons ($2, cons ($4, NULL)))),
	     NULL);
}
;

/* ids est une liste inversee d'identificateurs. */
ids : id { $$ = cons ($1, NULL); }
| ids ',' id { $$ = cons ($3, $1); }
;

type : id { $$ = $1; }
;

/* type_list est une liste inversee de types. */
type_list : { $$ = NULL; }
| non_empty_type_list { $$ = $1; }
;

/* non_empty_type_list est une liste inversee de types. */
non_empty_type_list : type { $$ = cons ($1, NULL); }
| non_empty_type_list ',' type { $$ = cons ($3, $1); }
;

optional_hash_keyword : kw_hash { $$ = 1; }
| { $$ = 0; }
;

block : '{' messages '}' { $$ = CALL ($2, apply (T_STOP, NULL)); }
;

messages : { $$ = newfun (do_identity, NULL); }
| messages message { $$ = compose ($1, $2); }
;

message : label id kw_arrow id ':' term
{
  cell *commrest = cons ($1, cons ($2, cons ($4, cons ($6, NULL))));
  cell *linfo = newlineinfo (&@1, &@6, NULL);
  $$ = newfun (do_comm, cons (linfo, commrest));
}
| id ':' id equals_or_assign_operator term
{
  cell *rest = cons ($1, cons ($3, cons ($5, NULL)));
  cell *linfo = newlineinfo (&@1, &@5, NULL);
  $$ = newfun (do_guard, cons (newint ($4), cons (linfo, rest)));
}
| '{' messages '}' { $$ = $2 }
| kw_switch term '{' cases '}'
{
  cell *linfo = newlineinfo (&@1, &@5, NULL);
  $$ = newfun (do_switch, cons ($2, cons ($4, linfo)));
}
;

label : kw_label { $$ = newstring (yytext); }
;

equals_or_assign_operator : kw_tst_equal { $$ = T_TST_EQUAL; }
| kw_assign { $$ = T_ASSIGN; }
;

/* cases est une liste inversee de cas. */
cases : { $$ = NULL; }
| cases case { $$ = cons ($2, $1); }
;

case : kw_case term ':' messages
{
  cell *linfo = newlineinfo (&@1, &@4, NULL);
  $$ = newfun (do_case, cons ($2, cons ($4, linfo)));
}
;

term : tuple { $$ = mktuple (&@1, &@1, $1); }
;

/* tuple est une liste inversee de termes. */
tuple : atomic_term_or_ciphertext { $$ = cons ($1, NULL); }
| tuple ',' atomic_term_or_ciphertext { $$ = cons ($3, $1); }
;

/* term_list est une liste inversee de terms. */
term_list : '(' ')' { $$ = NULL; }
| '(' tuple ')' { $$ = $2; }
;

atomic_term_or_ciphertext : atomic_term { $$ = $1; }
| '{' term '}' '_' atomic_term optional_encryption_algorithm
{
  $$ = newlineinfo (&@1, &@6, apply (T_CRYPT,
				     cons ($2, cons ($5, cons ($6, NULL)))));
}
| '[' term ']' '_' atomic_term optional_encryption_algorithm
{
  $$ = newlineinfo (&@1, &@6, apply (T_SIGN,
				     cons ($2, cons ($5, cons ($6, NULL)))));
}
| atomic_term '%' atomic_term
{
  $$ = newlineinfo (&@1, &@3, apply (T_PERCENT,
				     cons ($1, cons ($3, NULL))));
}
;

atomic_term : id term_list
{
  $$ = newlineinfo (&@1, &@2, apply (T_APP,
				     cons ($1, reverse ($2, NULL))));
}
| id { $$ = $1; }
| '(' term ')' { $$ = $2; }
| '(' ')'
{
  $$ = newlineinfo (&@1, &@2, apply (T_TUPLE, NULL));
}
;

optional_encryption_algorithm : { $$ = apply (T_VANILLA, NULL); }
| '^' atomic_term { $$ = $2; }
;

/* environment est une liste inversee de sessions. */
environment : { $$ = NULL; }
| environment session { $$ = cons ($2, $1); }
;

session : kw_session assignments
{
  $$ = newlineinfo (&@1, &@2, apply (T_SESSION, reverse ($2, NULL)));
}
;

/* assignments est une liste inversee d'affectations de noms de principaux a des roles. */
assignments : assignment { $$ = cons ($1, NULL); }
| assignments ',' assignment { $$ = cons ($3, $1); }
;

assignment : id ':' id
{
  $$ = newlineinfo (&@1, &@3, apply (T_ROLE, cons ($1, cons ($3, NULL))));
}
;

/* properties est une liste inversee de proprietes a verifier. */
properties : { $$ = NULL; }
| properties property { $$ = cons ($2, $1); }
;

property : id term_list
{
  $$ = newlineinfo (&@1, &@2, apply (T_CLAIM, cons ($1, reverse ($2, NULL))));
}
;

%%

static cell *newlineinfo (struct yyltype *start, struct yyltype *end, cell *c)
  {
    return apply (T_LINE,
		  cons (c,
			cons (newint (start->first_line),
			      cons (newint (start->first_column),
				    cons (newint (end->last_line),
					  cons (newint (end->last_column),
						cons (newstring (end->text?end->text:""),
						      NULL)))))));
  }

static void usage (void)
{
  fprintf (stderr, "evaparse [input-file].\n");
}

void print_string1 (FILE *f, char *s, int len)
{
  char *send = s+len;
  char c;

  for (; s<send; )
    {
      c = *s++;
      if ((c & 0x60)!=0)
	{
	  if (c=='\\' || c=='\"')
	    fputc ('\\', f);
	  fputc (c, f);
	}
      else switch (c)
	{
	case '\n': fputc ('\\', f); fputc ('n', f); break;
	case '\t': fputc ('\\', f); fputc ('t', f); break;
#if '\n'!='\r'
	case '\r': fputc ('\\', f); fputc ('r', f); break;
#endif
	case '\b': fputc ('\\', f); fputc ('b', f); break;
	case '\f': fputc ('\\', f); fputc ('f', f); break;
	default: fputc ('\\', f);
	  fputc (((c >> 6) & 7)+'0', f);
	  fputc (((c >> 3) & 7)+'0', f);
	  fputc (((c >> 0) & 7)+'0', f);
	  break;
	}
    }
}

void print_string (FILE *f, char *s, int len)
{
  fputc ('\"', f);
  print_string1 (f, s, len);
  fputc ('\"', f);
}

static void pcell (FILE *f, cell *c)
{
  if (c==NULL)
    fputs ("(nil)", f);
  else switch (c->type)
    {
    case X_FUNC:
      fputs ("<fun>", f);
      break;
    case X_INT:
      fprintf (f, "%d", c->what.i);
      break;
    case X_STRING:
      print_string (f, c->what.s.text, strlen (c->what.s.text));
      break;
    default: /* case X_CONS: */
      fputc ('(', f);
      fputs (tags[c->what.c.car->what.i], f);
      for (; c = c->what.c.cdr, c!=NULL; )
	{
	  fputc (' ', f);
	  pcell (f, c->what.c.car);
	}
      fputc (')', f);
      break;
    }
}

extern FILE *yyin;

int main (int argc, char *argv[])
{
  int i;
  char *infname = NULL;

  yyin = stdin;
  for (i=1; i<argc; i++)
    {
      switch (argv[i][0])
	{
	case '-':
	  usage ();
	  exit (1);
	  break;
	default:
	  if (infname!=NULL)
	    {
	      fprintf (stderr,
		       "evaparse: input file already specified ('%s').\n",
		       infname);
	      exit (1);
	    }
	  yyin = fopen (infname = argv[i], "r");
	  break;
	}
    }
  if (setjmp (errbuf)!=0)
    exit (1);
  (void) yyparse ();
  pcell (stdout, theReturn);
  fputs ("\n", stdout);
  fflush (stdout);
  return 0;
}
