%{

(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *		This was developed as part of the EVA project; as such, access
 *		to this software by partners of the RNTL EVA Project (Trusted
 *		Logic S.A., Verimag) is defined by the intellectual property
 *		agreement signed as part of this project.
 *)

open "lisptoken_h"
open "lispin_h"

fun yyerr (yyd, exc) = (hlex_begin (yyd, 0);
                        hlex_flush (yyd, hlex_current_buffer yyd);
                        raise exc);

fun read_octal (s, i) =
    let val n = size s
	val ir = ref i
        val acc = ref 0
    in
       while !ir<n do
         (acc := lsl (!acc, 3) + ordof (s, !ir)-48;
          inc ir);
       !acc
    end;

fun read_dec (s, i) =
    let val n = size s
	val ir = ref i
        val acc = ref 0
    in
       while !ir<n do
         (acc := !acc*10 + ordof (s, !ir)-48;
          inc ir);
       !acc
    end;

%}

%x STRING

SPACE [\ \t\n\r]+
DIG [0-9]+
ID [^" \t\n\r()]+

%%

	val string_buf = outstring "";

<INITIAL>{
{SPACE}	continue
\(	return LISP_LPAREN
\)	return LISP_RPAREN
{DIG}	return (LISP_INT (read_dec (hlex_text yyd, 0)))
{ID}	return (LISP_ID (hlex_text yyd))
\"	#seek string_buf 0; #truncate string_buf (); hlex_begin (yyd, STRING); continue
<<EOF>>	yyerr (yyd, LispEof)
}

<STRING>{
\"	hlex_begin (yyd, INITIAL); return (LISP_STRING (#convert string_buf ()))
\n+	continue
\\[0-7]{1,3}	#put string_buf (chr (read_octal (hlex_text yyd, 1))); continue
\\[0-7]+	yyerr (yyd, LispBadEscape (hlex_text yyd))
\\n	#put string_buf "\n"; continue
\\t	#put string_buf "\t"; continue
\\r	#put string_buf "\^M"; continue
\\b	#put string_buf "\^G"; continue
\\f	#put string_buf "\f"; continue
\\.	#put string_buf (chr (ordof (hlex_text yyd, 1))); continue
[^\\\n\"]+	#put string_buf (hlex_text yyd); continue
<<EOF>>	yyerr (yyd, LispEofInString)
.	#put string_buf (hlex_text yyd); continue
}

%%

exception LispNoFunction
exception LispUnexpectedFunction of string
exception LispRParenEvt
exception LispExpectedInt
exception LispExpectedFile
exception LispExpectedRParen

fun lispread1 yyd =
    (case yylex yyd of
	 LISP_LPAREN => (case yylex yyd of
			     LISP_ID "line" =>
			     let val t = lispread1 yyd
				 val m = lispreadint yyd
				 val n = lispreadint yyd
				 val p = lispreadint yyd
				 val q = lispreadint yyd
				 val s = lispreadstring yyd
			     in
				case yylex yyd of
				    LISP_RPAREN =>
				    LINE (t, |[ first_line = m,
						first_column = n,
						last_line = p,
						last_column = q,
						in_file = s
						]|)
				  | _ => raise LispExpectedRParen
			     end
			   | LISP_ID fname => APP (fname, lispreadlist yyd)
			   | _ => raise LispNoFunction)
       | LISP_RPAREN => raise LispRParenEvt
       | LISP_INT n => INT n
       | LISP_STRING s => VAR s
       | LISP_ID s => raise (LispUnexpectedFunction s))
and lispreadlist yyd =
    (lispread1 yyd::lispreadlist yyd handle LispRParenEvt => nil)
and lispreadint yyd =
    (case lispread1 yyd of
	 INT n => n
       | _ => raise LispExpectedInt)
and lispreadstring yyd =
    (case lispread1 yyd of
	 VAR s => s
       | APP ("nil", []) => ""
       | _ => raise LispExpectedFile);

fun lispread f =
    let val yyd = hlex_data (f, fn _ => true)
    in
	lispread1 yyd
    end;

fun lispwrite (f as |[put, ...]|) =
    let fun outint i = print f (pack (i:int))
	fun outintsp i = (put " "; outint i)
	fun outescape s =
	    iterate
	      if i=34
		  then put "\\\""
	      else if i<32 orelse i>=128 andalso i<160
		  then let val (a, i3) = i divmod 8
			   val (i1, i2) = a divmod 8
		       in
			   put "\\";
			   put (chr (i1+48));
			   put (chr (i2+48));
			   put (chr (i3+48))
		       end
	      else put c
	    | c in list explode s
	    val i = ord c
	    end
	val margin = ref 0
	fun cr () = let val i = ref (!margin)
		    in
			put "\n";
			while (!i<>0) do
			    (dec i;
			     put " ")
		    end
	fun write (VAR s) = (put "\""; outescape s; put "\"")
	  | write (INT i) = outint i
	  | write (APP (f, l)) = (put "("; put f; wl l; put ")")
	  | write (LINE (t, |[ first_line, first_column, last_line, last_column, in_file ]|)) =
	    (put "(line "; write t;
	     outintsp first_line; outintsp first_column; outintsp last_line; outintsp last_column;
	     put " \""; outescape in_file; put "\")")
        and wl l = (margin := !margin + 2;
		    iterate
		      (cr (); write t)
		    | t in list l
		    end;
		    margin := !margin - 2)
    in
	fn t => (margin := 0; write t)
    end;

